"""
Get the numerical vs the theoretical coverage for different coverage factors
and plot them. This module contains
- The function `get_coverages`, that returns two numpy arrays containing
  the numerical and theoretical coverage.
- The function `get_coverages_with_uncertainties`, that runs through several
  networks and collects the results of `get_coverages`.
- The function `plot_coverages`, that plots the results of
  `get_coverage_distribution`.
"""
import importlib
import os
import argparse
import json

import numpy as np
import torch
import torch.backends.cudnn
from torch.utils.data import DataLoader
from tqdm import tqdm

from EIVArchitectures import Networks
from EIVTrainingRoutines import train_and_store
from EIVGeneral.coverage_metrics import epistemic_coverage, normalized_std
from EIVData.repeated_sampling import repeated_sampling

import matplotlib.pyplot as plt

def get_coverages(not_averaged_predictions, y,\
        q_range=np.linspace(0.1,0.9,num=30)):
    """
    Compute the numerical and theoretical coverage for various coverage
    factors, computed from the `q` in `q_range`, using
    `EIVGeneral.epistemic_coverage` 
    :param not_averaged_predictions: A tuple of tensors as in the output of
    `FNNEIV.predict` with `take_average_of_prediction` set to `False`, i.e.:
    the predictions of the neural net not averaged over the first dimension 
    (the repetition dimension in `FNNEIV.predict`) and
    the aleatoric uncertainty with a batch dimension and a feature dimension. 
    :param y: A `torch.tensor` of the same shape then the second components
    of `not_averaged_predictions`. If the feature dimension is missing, it is added.
    :param q_range: An iterator through floats between 0 and 1.
    :returns: numerical_coverage, theoretical_coverage (numpy arrays)
    """
    numerical_coverage_collection = []
    theoretical_coverage_collection = []
    for q in q_range:
        numerical_coverage, theoretical_coverage =\
                epistemic_coverage(\
                not_averaged_predictions=not_averaged_predictions,
                y=y, q=q)
        numerical_coverage_collection.append(float(numerical_coverage))
        theoretical_coverage_collection.append(float(theoretical_coverage))
    return np.array(numerical_coverage_collection),\
            np.array(theoretical_coverage_collection)
    

def get_coverage_distribution(net_iterator, dataloader_iterator,
        device, number_of_draws, q_range=np.linspace(0.1,0.9,num=30),
        number_of_test_samples=10, stack=True):
    """
    Returns the numerical and theoretical coverage for all nets in
    `net_iterator` with dataloader from `dataloader_iterator`
    :param net_iterator: Iterator yielding `torch.nn.Module`
    :param dataloader_iterator: Iterator yielding `torch.utils.data.DataLoader`.
    Should return the same number of elements than `net_iterator`.
    :param device: The `torch.device` to be used.
    :param number_of_draws: Integer, to be used as keyword for all networks
    returned by `net_iterator`. Defaults to 10.
    :param q_range: Range of floats between 0 and 1, as in 
    `EIVGeneral.coverage_plot.get_coverages`.
    :param number_of_test_samples: An integer, number of samples from the
    dataloaders in dataloader_iterator to be used.
    :param stack: Boolean. If True (default) the results will be stacked along 
    the last dimension. If False a list will be returned.
    :returns: num_coverage_collection, th_coverage_collection
    """
    num_coverage_collection, th_coverage_collection = [],[]
    for net, dataloader in\
     zip(net_iterator, dataloader_iterator):
                not_av_pred_collection_out, not_av_pred_collection_sigma,\
                    y_collection = [], [], []
                for i, (x,y) in enumerate(dataloader):
                    x, y = x.to(device), y.to(device)
                    not_averaged_predictions = net.predict(x,
                            take_average_of_prediction=False,
                            number_of_draws=number_of_draws)
                    not_av_pred_collection_out.append(
                            not_averaged_predictions[0])
                    not_av_pred_collection_sigma.append(
                            not_averaged_predictions[1])
                    y_collection.append(y)
                not_av_pred_collection = [
                        torch.concat(not_av_pred_collection_out, dim=0), 
                        torch.concat(not_av_pred_collection_sigma, dim=0)]
                y_collection = torch.concat(y_collection, dim=0)
                numerical_coverage, theoretical_coverage = get_coverages(
                        not_averaged_predictions=not_av_pred_collection,
                        y=y_collection)
                num_coverage_collection.append(numerical_coverage)
                th_coverage_collection.append(theoretical_coverage)
    if stack:
        num_coverage_collection = np.stack(num_coverage_collection, axis=-1)
        th_coverage_collection = np.stack(th_coverage_collection, axis=-1)
    return num_coverage_collection, th_coverage_collection

                





#######

# data = 'linear'
# # load hyperparameters from JSON file
# with open(os.path.join('/home/martin09/san/Projects/journal_eiv/Experiments/configurations',f'eiv_{data}.json'),'r') as conf_file:
#     eiv_conf_dict = json.load(conf_file)
# with open(os.path.join('/home/martin09/san/Projects/journal_eiv/Experiments/configurations',f'noneiv_{data}.json'),'r') as conf_file:
#     noneiv_conf_dict = json.load(conf_file)
# seed = 0
# long_dataname = eiv_conf_dict["long_dataname"]
# short_dataname = eiv_conf_dict["short_dataname"]
# load_data = importlib.import_module(f'EIVData.{long_dataname}').load_data
# train, test = load_data()
# test_dataloader = DataLoader(test, batch_size=1000)
# device = torch.device('cuda:0')
# x,y = next(iter(test_dataloader))
# if len(y.shape) <= 1:
#     y = y.view((-1,1))
# if len(x.shape) <= 1:
#     x = x.view((-1,1))
# x,y = x.to(device), y.to(device)
# input_dim = x.shape[1]
# output_dim = y.shape[1]
# init_std_y = noneiv_conf_dict["init_std_y_list"][0]
# unscaled_reg = noneiv_conf_dict["unscaled_reg"]
# p = noneiv_conf_dict["p"]
# hidden_layers = noneiv_conf_dict["hidden_layers"]
# saved_file = os.path.join('saved_networks',
#             f'noneiv_{short_dataname}'\
#                     f'_init_std_y_{init_std_y:.3f}_ureg_{unscaled_reg:.1f}'\
#                     f'_p_{p:.2f}_seed_{seed}.pkl')
# noneiv_net = Networks.FNNBer(p=p, init_std_y=init_std_y,
#         h=[input_dim, *hidden_layers, output_dim]).to(device)
# train_and_store.open_stored_training(saved_file=saved_file,
#         net=noneiv_net, device=device)
# # EiV
# init_std_y = eiv_conf_dict["init_std_y_list"][0]
# unscaled_reg = eiv_conf_dict["unscaled_reg"]
# p = eiv_conf_dict["p"]
# hidden_layers = eiv_conf_dict["hidden_layers"]
# fixed_std_x = eiv_conf_dict["fixed_std_x"]
# saved_file = os.path.join('saved_networks',
#         f'eiv_{short_dataname}'\
#                 f'_init_std_y_{init_std_y:.3f}_ureg_{unscaled_reg:.1f}'\
#                 f'_p_{p:.2f}_fixed_std_x_{fixed_std_x:.3f}'\
#                 f'_seed_{seed}.pkl')
# eiv_net = Networks.FNNEIV(p=p, init_std_y=init_std_y,
#         h=[input_dim, *hidden_layers, output_dim],
#         fixed_std_x=fixed_std_x).to(device)
# def eiv_net_iterator(seed_range=range(0,10)):

# train_and_store.open_stored_training(saved_file=saved_file,
#         net=eiv_net, device=device)

# noneiv_not_averaged_predictions = noneiv_net.predict(x,\
#         number_of_draws=100, 
#         take_average_of_prediction=False)
# noneiv_num_cov, noneiv_th_cov = get_coverages(noneiv_not_averaged_predictions, y,\
#         q_range=np.linspace(0.01,0.99,num=30))
# eiv_not_averaged_predictions = eiv_net.predict(x,\
#         number_of_draws=[100,5], 
#         take_average_of_prediction=False)
# eiv_num_cov, eiv_th_cov = get_coverages(eiv_not_averaged_predictions, y,\
#         q_range=np.linspace(0.01,0.99,num=30))
# lin_x = np.linspace(np.min(noneiv_th_cov), np.max(noneiv_th_cov))
# plt.plot(lin_x, lin_x)
# plt.plot(noneiv_th_cov, noneiv_num_cov)
# plt.plot(eiv_th_cov, eiv_num_cov)
